/*****************************************************************************
 *
 * This file is part of Mapnik (c++ mapping toolkit)
 *
 * Copyright (C) 2024 Artem Pavlenko
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *****************************************************************************/
#include <mapnik/warning.hpp>
// clang-format off
MAPNIK_DISABLE_WARNING_UNKNOWN_PRAGMAS // clang+gcc         
MAPNIK_DISABLE_WARNING_PRAGMAS // gcc              
MAPNIK_DISABLE_WARNING_PRAGMA_MESSAGES         
MAPNIK_DISABLE_WARNING_UNSEQUENCED             
MAPNIK_DISABLE_WARNING_UNUSED_FUNCTION         
MAPNIK_DISABLE_WARNING_UNUSED_PARAMETER        
MAPNIK_DISABLE_WARNING_REDECLARED_CLASS_MEMBER 
MAPNIK_DISABLE_WARNING_UNUSED_LOCAL_TYPEDEF    
MAPNIK_DISABLE_WARNING_UNUSED_LOCAL_TYPEDEFS // gcc5
MAPNIK_DISABLE_WARNING_SHADOW                  
MAPNIK_DISABLE_WARNING_CPP11_NARROWING         
MAPNIK_DISABLE_WARNING_SIGN_CONVERSION         
MAPNIK_DISABLE_WARNING_SIGN_COMPARE            
MAPNIK_DISABLE_WARNING_CONVERSION              
MAPNIK_DISABLE_WARNING_CPP_98_COMPAT_PEDANTIC  
MAPNIK_DISABLE_WARNING_DISABLE_MACRO_EXPANSION 
MAPNIK_DISABLE_WARNING_OLD_STYLE_CAST          
MAPNIK_DISABLE_WARNING_DOCUMENTATION           
MAPNIK_DISABLE_WARNING_DOCUMENTATION_UNKNOWN_COMMAND
MAPNIK_DISABLE_WARNING_UNDEF                  
MAPNIK_DISABLE_WARNING_DEPRECATED             
MAPNIK_DISABLE_WARNING_PADDED                 
MAPNIK_DISABLE_WARNING_CPP_98_COMPAT          
MAPNIK_DISABLE_WARNING_RESERVED_ID_MACRO      
MAPNIK_DISABLE_WARNING_WEAK_VTABLE            
MAPNIK_DISABLE_WARNING_EXTRA_SEMI             
MAPNIK_DISABLE_WARNING_GLOBAL_CONSTRUCTORS    
MAPNIK_DISABLE_WARNING_HEADER_HYGIENE         
MAPNIK_DISABLE_WARNING_EXIT_TIME_DESTRUCTORS  
MAPNIK_DISABLE_WARNING_SWITCH_ENUM            
MAPNIK_DISABLE_WARNING_MISSING_NORETURN       
MAPNIK_DISABLE_WARNING_COVERED_SWITCH_DEFAULT 
MAPNIK_DISABLE_WARNING_FLOAT_EQUAL
  // clang-format on
